#!/usr/bin/env python3
'''
Program to rename, remove, or copy files and directories using your
editor. Will use git to action the rename and remove if run within a git
repository.
'''
# Author: Mark Blakeney, May 2019.

import argparse
import itertools
import os
import re
import shlex
import shutil
import subprocess
import sys
import tempfile
from collections import OrderedDict
from pathlib import Path
from typing import List, Optional, TextIO, Tuple

from platformdirs import user_config_path

# Some constants
PROG = Path(sys.argv[0]).stem
CNFFILE = user_config_path() / f'{PROG}-flags.conf'
EDITOR = PROG.upper() + '_EDITOR'
SUFFIX = '.sh'

# The temp dir we will use in the dir of each target move
TEMPDIR = '.tmp-' + PROG

# Define ANSI escape sequences for colors ..
COLOR_red = '\033[31m'
COLOR_green = '\033[32m'
COLOR_yellow = '\033[33m'
COLOR_reset = '\033[39m'

ACTIONS = {
    'remove': (COLOR_red, ('Removing', 'Remove', 'Removed')),
    'rename': (COLOR_yellow, ('Renaming', 'Rename', 'Renamed')),
    'copy': (COLOR_green, ('Copying', 'Copy', 'Copied')),
}

args = argparse.Namespace()
gitfiles = set()
counts = [0, 0]

EDITORS = {'Windows': 'notepad', 'Darwin': 'open -e', 'default': 'vim'}

def get_default_editor() -> str:
    'Return default editor for this system'
    from platform import system
    return EDITORS.get(system(), EDITORS['default'])

def log(action: str, msg: str, error: Optional[str] = None, *,
        prompt: bool = False) -> None:
    'Output message with appropriate color'
    counts[bool(error)] += 1

    color, tense = ACTIONS[action]

    if error:
        out = sys.stderr
        msg = f'{tense[1]} {msg} ERROR: {error}'
    elif args.quiet and not prompt:
        return
    else:
        out = sys.stdout
        msg = f'{tense[0] if prompt else tense[2]} {msg}'

    if args.no_color:
        print(msg, file=out)
    else:
        print(color + msg + COLOR_reset, file=out)

def run(cmd: str) -> Tuple[str, str]:
    'Run given command and return (stdout, stderr) strings'
    stdout = ''
    stderr = ''
    try:
        res = subprocess.run(cmd, stdout=subprocess.PIPE, shell=True,
                             stderr=subprocess.PIPE, universal_newlines=True)
    except Exception as e:
        stderr = str(e)
    else:
        if res.stdout:
            stdout = res.stdout.strip()
        if res.stderr:
            stderr = res.stderr.strip()

    return stdout, stderr

def remove(path: Path, git: bool = False, trash: bool = False,
           recurse: bool = False) -> Optional[str]:
    'Delete given file/directory'
    if not recurse and not path.is_symlink() and path.is_dir() and \
            any(path.iterdir()):
        return 'Directory not empty'

    if git:
        ropt = '-r' if recurse else ''
        out, err = run(f'git rm -f {ropt} "{path}"')
        return f'git error: {err}' if err else None

    if trash:
        out, err = run(f'{args.trash_program} "{path}"')
        return f'{args.trash_program} error: {err}' if err else None

    if recurse:
        try:
            shutil.rmtree(str(path))
        except Exception as e:
            return str(e)
    else:
        try:
            if not path.is_symlink() and path.is_dir():
                path.rmdir()
            else:
                path.unlink()
        except Exception as e:
            return str(e)

def rename(pathsrc: Path, pathdest: Path, is_git: bool = False) -> None:
    'Rename given pathsrc to pathdest'
    if is_git:
        out, err = run(f'git mv -f "{pathsrc}" "{pathdest}"')
        if err:
            log('rename', '"{pathsrc}"', f'git mv error: {err}')
    else:
        shutil.move(str(pathsrc), str(pathdest))

def create_options(options: str) -> str:
    'Create a string of options + keys for user to choose from'
    letters = []
    words = []
    for arg in options.split():
        letters.append(arg[0])
        words.append(f'({arg[0].upper()}){arg[1:]}')
    return ', '.join(words) + ': [' + '/'.join(letters) + ']? '

class Fpath:
    'Class to manage each instance of a file/dir'
    paths = []
    tempdirs = set()

    def __init__(self, path: Path):
        'Class constructor'
        self.path = path
        self.newpath = None
        self.temppath = None
        self.note = ''
        self.copies: List[Path] = []
        self.is_dir = path.is_dir()
        self.appdash = '/' if self.is_dir else ''
        self.diagrepr = str(self.path)
        self.is_git = self.diagrepr in gitfiles

        self.linerepr = self.diagrepr if self.diagrepr.startswith('/') \
                else './' + self.diagrepr
        if self.is_dir and not self.diagrepr.endswith('/'):
            self.linerepr += '/'
            self.diagrepr += '/'

    @staticmethod
    def inc_path(path: Path) -> Optional[Path]:
        'Find next unique file name'
        # Iterate forever, there can only be a finite number of existing
        # paths
        name = path.name
        for c in itertools.count():
            if not path.is_symlink() and not path.exists():
                return path
            path = path.with_name(name + ('~' if c <= 0 else f'~{c}'))

    def copy(self, pathdest: Path) -> Optional[str]:
        'Copy given pathsrc to pathdest'
        func = shutil.copytree if self.is_dir else shutil.copy2
        try:
            func(str(self.newpath), str(pathdest))  # type:ignore
        except Exception as e:
            return str(e)

    def rename_temp(self) -> None:
        'Move this path to a temp place in advance of final move'
        if not self.newpath:
            return
        tempdir = self.newpath.parent / TEMPDIR
        try:
            tempdir.mkdir(parents=True, exist_ok=True)
        except Exception:
            print(f'Create dir for {self.diagrepr} ERROR: '
                    f'Can not write in {tempdir.parent}', file=sys.stderr)
        else:
            self.temppath = self.inc_path(tempdir / self.newpath.name)
            if self.temppath:
                self.tempdirs.add(tempdir)
                rename(self.path, self.temppath, self.is_git)

    def restore_temp(self) -> bool:
        'Restore temp path to final destination'
        if not self.temppath or not self.newpath:
            return False
        self.newpath = self.inc_path(self.newpath)
        if self.newpath:
            rename(self.temppath, self.newpath, self.is_git)
            return True
        return False

    def sort_name(self) -> str:
        'Return name for sort'
        return str(self.path)

    def sort_time(self) -> float:
        'Return time for sort'
        return self.path.stat().st_mtime

    def sort_size(self) -> int:
        'Return size for sort'
        return self.path.stat().st_size

    def is_recursive(self) -> bool:
        'Return True if directory and we can view it and contains children'
        if not self.is_dir:
            return False
        try:
            return any(self.path.iterdir())
        except Exception:
            return False

    def log_pending_changes(self) -> None:
        'Log all pending changes for this path'
        if not self.newpath:
            log('remove', f'"{self.diagrepr}"{self.note}', prompt=True)
        elif self.newpath != self.path:
            log('rename', f'"{self.diagrepr}" to '
                f'"{self.newpath}{self.appdash}"', prompt=True)
        for c in self.copies:
            log('copy', f'"{self.diagrepr}" to '
                f'"{c}{self.appdash}"{self.note}', prompt=True)

    @classmethod
    def remove_temps(cls) -> None:
        'Remove all the temp dirs we created in rename_temp() above'
        for p in cls.tempdirs:
            remove(p, git=False, trash=False, recurse=True)

        cls.tempdirs.clear()

    @classmethod
    def append(cls, path: Path) -> None:
        'Add a single file/dir to the list of paths'
        # Filter out files/dirs if asked
        if args.files:
            if path.is_dir():
                return
        elif args.dirs:
            if not path.is_dir():
                return

        # Filter out links if asked
        if args.nolinks and path.is_symlink():
            return

        cls.paths.append(cls(path))

    @classmethod
    def add(cls, name: str, expand: bool) -> None:
        'Add file[s]/dir[s] to the list of paths'
        path = Path(name)
        if not path.exists():
            sys.exit(f'ERROR: {name} does not exist')

        if expand and path.is_dir():
            for child in sorted(path.iterdir()):
                if args.all or not child.name.startswith('.'):
                    cls.append(child)
        else:
            cls.append(path)

    @classmethod
    def writefile(cls, fp: TextIO) -> None:
        'Write the file for user to edit'
        # Ensure consistent width for line numbers
        num_width = len(str(len(cls.paths)))
        fp.writelines(f'{i:0{num_width}}  {p.linerepr}\n'
                      for i, p in enumerate(cls.paths, 1))

    @classmethod
    def readfile(cls, fp: TextIO) -> None:
        'Read the list of files/dirs as edited by user'
        for count, line in enumerate(fp, 1):
            # Skip blank or commented lines
            rawline = line.rstrip('\n\r')
            line = rawline.lstrip()
            if not line or line[0] == '#':
                continue

            try:
                n, pathstr = line.split(maxsplit=1)
            except Exception:
                sys.exit(f'ERROR: line {count} invalid:\n{rawline}')
            try:
                num = int(n)
            except Exception:
                sys.exit(f'ERROR: line {count} number {n} invalid:\n{rawline}')

            if num <= 0 or num > len(cls.paths):
                sys.exit(f'ERROR: line {count} number {num} '
                        f'out of range:\n{rawline}')

            path = cls.paths[num - 1]

            if len(pathstr) > 1:
                pathstr = pathstr.rstrip('/')

            newpath = Path(pathstr)

            if path.newpath:
                if newpath != path.path:
                    path.copies.append(newpath)
            else:
                path.newpath = newpath

    @classmethod
    def reset(cls) -> None:
        'Reset all the read path changes to null state'
        for p in cls.paths:
            p.newpath = None
            p.copies.clear()

    @classmethod
    def get_path_changes(cls) -> List:
        'Get a list of change paths from the user'
        with tempfile.TemporaryDirectory() as fdir:
            # Create a temp file for the user to edit then read the lines back
            fpath = Path(fdir, f'{PROG}{args.suffix}')
            restart = True
            while True:
                cls.reset()
                if restart:
                    restart = False
                    with fpath.open('w') as fp:
                        cls.writefile(fp)

                # Invoke editor on list of paths
                editfile(fpath)

                # Read the changed paths
                with fpath.open() as fp:
                    cls.readfile(fp)

                # Reduce paths to those that were removed or changed by the user
                paths = [p for p in cls.paths
                         if p.path != p.newpath or p.copies]

                if not paths:
                    return []

                # Lazy eval the next path value
                for p in paths:
                    p.note = ' recursively' if p.is_recursive() else ''

                if not args.interactive:
                    return paths

                # Prompt user with pending changes if required
                for p in paths:
                    p.log_pending_changes()

                options = create_options('proceed edit restart quit')
                while True:
                    try:
                        ans = input(options).lower()
                    except KeyboardInterrupt:
                        print()
                        return []

                    if ans == 'p':
                        return paths
                    elif ans == 'e':
                        break
                    elif ans == 'r':
                        restart = True
                        break
                    elif ans == 'q':
                        return []
                    else:
                        print('Invalid answer.')

def editfile(filename: Path) -> None:
    'Run the editor command'
    # Use explicit user defined editor or choose system default
    editor = os.getenv(EDITOR) or os.getenv('EDITOR') or get_default_editor()
    editcmd = shlex.split(editor) + [str(filename)]

    # Run the editor ..
    with open('/dev/tty') as tty:
        res = subprocess.run(editcmd, stdin=tty)

    # Check if editor returned error
    if res.returncode != 0:
        sys.exit(f'ERROR: {editor} returned error {res.returncode}')

def main() -> int:
    'Main code'
    global args

    # Process command line options
    opt = argparse.ArgumentParser(description=__doc__.strip(),
            epilog='Note you can set default starting options in '
            f'{CNFFILE}. The negation options (i.e. the --no-* options '
            'and their shortforms) allow you to temporarily override your '
            'defaults.')
    opt.add_argument('-i', '--interactive', action='store_true',
            help='prompt with summary of changes before doing them')
    opt.add_argument('-I', '--no-interactive', dest='interactive',
                     action='store_false',
            help='negate the -i/--interactive option')
    opt.add_argument('-a', '--all', action='store_true',
            help='include all (including hidden) files')
    opt.add_argument('-A', '--no-all', dest='all', action='store_false',
            help='negate the -a/--all option')
    opt.add_argument('-r', '--recurse', action='store_true',
            help='recursively remove any files and directories in '
            'removed directories')
    opt.add_argument('-R', '--no-recurse', dest='recurse', action='store_false',
            help='negate the -r/--recurse option')
    opt.add_argument('-q', '--quiet', action='store_true',
            help='do not print successful rename/remove/copy actions')
    opt.add_argument('-Q', '--no-quiet', dest='quiet', action='store_false',
            help='negate the -q/--quiet option')
    opt.add_argument('-G', '--no-git', dest='git',
            action='store_const', const=0,
            help='do not use git if invoked within a git repository')
    opt.add_argument('-g', '--git', dest='git',
            action='store_const', const=1,
            help='negate the --no-git option and DO use automatic git')
    opt.add_argument('-t', '--trash', action='store_true',
            help='use trash program to do deletions')
    opt.add_argument('-T', '--no-trash', dest='trash', action='store_false',
            help='negate the -t/--trash option')
    opt.add_argument('--trash-program', default='trash-put',
            help='trash program to use, default="%(default)s"')
    opt.add_argument('-c', '--no-color', action='store_true',
            help='do not color rename/remove/copy messages')
    opt.add_argument('-d', '--dirnames', action='store_true',
            help='edit given directory names directly, not their contents')
    grp = opt.add_mutually_exclusive_group()
    grp.add_argument('-F', '--files', action='store_true',
            help='only show/edit files')
    grp.add_argument('-D', '--dirs', action='store_true',
            help='only show/edit directories')
    opt.add_argument('-L', '--nolinks', action='store_true',
            help='ignore all symlinks')
    opt.add_argument('-N', '--sort-name', dest='sort',
            action='store_const', const=1,
            help='sort paths in file by name, alphabetically')
    opt.add_argument('-M', '--sort-time', dest='sort',
            action='store_const', const=2,
            help='sort paths in file by time, oldest first')
    opt.add_argument('-S', '--sort-size', dest='sort',
            action='store_const', const=3,
            help='sort paths in file by size, smallest first')
    opt.add_argument('-E', '--sort-reverse', action='store_true',
            help='sort paths (by name/time/size) in reverse')
    opt.add_argument('-X', '--group-dirs-first', dest='group_dirs',
            action='store_const', const=1,
            help='group directories first (including when sorted)')
    opt.add_argument('-Y', '--group-dirs-last', dest='group_dirs',
            action='store_const', const=0,
            help='group directories last (including when sorted)')
    opt.add_argument('-Z', '--no-group-dirs', dest='group_dirs',
            action='store_const', const=-1,
            help='negate the options to group directories')
    opt.add_argument('--suffix', default=SUFFIX,
            help='specify suffix for editor file, default="%(default)s"')
    opt.add_argument('-V', '--version', action='store_true',
            help=f'show {PROG} version')
    opt.add_argument('args', nargs='*',
            help='file|dir, or "-" for stdin')

    # Merge in default args from user config file. Then parse the
    # command line.
    if CNFFILE.exists():
        with CNFFILE.open() as fp:
            lines = [re.sub(r'#.*$', '', line).strip() for line in fp]
        cnflines = ' '.join(lines).strip()
    else:
        cnflines = ''

    args = opt.parse_args(shlex.split(cnflines) + sys.argv[1:])

    if args.version:
        if sys.version_info >= (3, 8):
            from importlib.metadata import version
        else:
            from importlib_metadata import version

        try:
            ver = version(PROG)
        except Exception:
            ver = 'unknown'

        print(ver)
        return 0

    # Check if we are in a git repo
    if args.git != 0:
        out, giterr = run('git ls-files')
        if giterr and args.git:
            print(f'Git invocation error: {giterr}', file=sys.stderr)
        if out:
            gitfiles.update(out.splitlines())

        if args.git and not gitfiles:
            opt.error('must be within a git repo to use -g/--git option')

    # Set input list to a combination of arguments and stdin
    filelist = args.args
    if sys.stdin.isatty():
        if not filelist:
            filelist.append('.')
    elif '-' not in filelist:
        filelist.insert(0, '-')

    # Iterate over all (unique) inputs to get a list of files/dirs
    for name in OrderedDict.fromkeys(filelist):
        if name == '-':
            for line in sys.stdin:
                name = line.rstrip('\n\r')
                if name != '.':
                    Fpath.add(line.rstrip('\n\r'), False)
        else:
            Fpath.add(name, not args.dirnames)

    # Sanity check that we have something to edit
    if not Fpath.paths:
        desc = 'files' if args.files else \
                'directories' if args.dirs else 'files or directories'
        print(f'No {desc}.')
        return 0

    if args.sort == 1:
        Fpath.paths.sort(key=Fpath.sort_name, reverse=args.sort_reverse)
    elif args.sort == 2:
        Fpath.paths.sort(key=Fpath.sort_time, reverse=args.sort_reverse)
    elif args.sort == 3:
        Fpath.paths.sort(key=Fpath.sort_size, reverse=args.sort_reverse)

    if args.group_dirs is not None and args.group_dirs >= 0:
        ldirs: List[Fpath] = []
        lfiles: List[Fpath] = []
        for path in Fpath.paths:
            (ldirs if path.is_dir else lfiles).append(path)
        Fpath.paths = ldirs + lfiles if args.group_dirs else lfiles + ldirs

    paths = Fpath.get_path_changes()
    if not paths:
        return 0

    err: Optional[str]

    # Pass 1: Rename all moved files & dirs to temps, delete all removed
    # files.
    for p in paths:
        if p.newpath:
            if p.newpath != p.path:
                p.rename_temp()
        elif not p.is_dir:
            err = remove(p.path, p.is_git, args.trash)
            log('remove', f'"{p.diagrepr}"', err)

    # Pass 2: Delete all removed dirs, if empty or recursive delete.
    for p in paths:
        if p.is_dir and not p.newpath:
            if remove(p.path, p.is_git, args.trash, args.recurse) is None:
                # Have removed, so flag as finished for final dirs pass below
                p.is_dir = False
                log('remove', f'"{p.diagrepr}"{p.note}')

    # Pass 3. Rename all temp files and dirs to final target, and make
    # copies.
    for p in paths:
        if p.restore_temp():
            log('rename', f'"{p.diagrepr}" to "{p.newpath}{p.appdash}"')

        for c in p.copies:
            err = p.copy(c)
            log('copy', f'"{p.diagrepr}" to "{c}{p.appdash}"{p.note}', err)

    # Remove all the temporary dirs we created
    Fpath.remove_temps()

    # Pass 4. Delete all remaining dirs
    for p in paths:
        if p.is_dir and not p.newpath:
            err = remove(p.path, p.is_git, args.trash, args.recurse)
            log('remove', f'"{p.diagrepr}"{p.note}', err)

    # Return status code 0 = all good, 1 = some bad, 2 = all bad.
    return (1 if counts[0] > 0 else 2) if counts[1] > 0 else 0

if __name__ == '__main__':
    sys.exit(main())
